/************************************************************************
*
Legrand MyHome / OpenWebNet USB Interface board driver for Domoticz
Date: 05-10-2016
Written by: St�phane Lebrasseur
License: Public domain
************************************************************************/

#include "stdafx.h"
#include "OpenWebNetUSB.h"
#include "openwebnet/bt_openwebnet.h"
#include "../main/Helper.h"
#include "../main/Logger.h"
#include "../main/RFXtrx.h"
#include "P1MeterBase.h"
#include "hardwaretypes.h"
#include "../main/SQLHelper.h"

#include <algorithm>
#include <ctime>
#include <boost/exception/diagnostic_information.hpp>
#include <iostream>
#include <string>

COpenWebNetUSB::COpenWebNetUSB(const int ID, const std::string& devname, unsigned int baud_rate)
{
	m_HwdID = ID;
	m_szSerialPort = devname;
	m_iBaudRate = baud_rate;
	m_retrycntr = RETRY_DELAY - 2;
	m_bHaveReceived = false;
	m_bEnableReceive = true;
}

bool COpenWebNetUSB::StartHardware()
{
	RequestStart();

	m_retrycntr = RETRY_DELAY - 2; //will force reconnect first thing

								   //Start worker thread
	m_thread = std::make_shared<std::thread>([this] { Do_Work(); });
	SetThreadNameInt(m_thread->native_handle());
	return (m_thread != nullptr);
}

bool COpenWebNetUSB::StopHardware()
{
	if (m_thread)
	{
		RequestStop();
		m_thread->join();
		m_thread.reset();
	}
	m_bIsStarted = false;
	return true;
}

void COpenWebNetUSB::Do_Work()
{
	while (!IsStopRequested(OPENWEBNET_HEARTBEAT_DELAY))
	{
		m_LastHeartbeat = mytime(nullptr);
	}
	terminate();

	Log(LOG_STATUS, "Heartbeat worker stopped...");
}

/**
Convert domoticz command in a OpenWebNet command, then send it to device
**/
bool COpenWebNetUSB::WriteToHardware(const char *pdata, const unsigned char length)
{
	_tGeneralSwitch *pCmd = (_tGeneralSwitch*)pdata;

	unsigned char packetlength = pCmd->len;
	unsigned char packettype = pCmd->type;
	unsigned char subtype = pCmd->subtype;

	int who = 0;
	int what = 0;
	std::stringstream whereStr;
	std::stringstream devIdStr;

	switch (subtype) {
		case sSwitchBlindsT1:
		case sSwitchLightT1:
			//Bus command
			whereStr << (int)(pCmd->id & 0xFFFF);
			break;
		case sSwitchBlindsT2:
		case sSwitchLightT2:
			//Zigbee command
			//Transmission unicast (no prefix) in ZigBee (#9 suffix) , command light ON and ID = 0x000501F8 (0d328184) and you want to control the UNIT 2 : * 1 * 1 * 32818402#9##
			//Transmission unicast (no prefix) in ZigBee (#9 suffix) , command light ON and ID = 0x000501F8 (0d328184) and you want to control the all UNIT : * 1 * 1 * 32818400#9##
			//Transmission broadcast (0# prefix) in ZigBee (#9 suffix), command light OFF and you want to control the UNIT 1 : * 1 * 0 * 0#01#9##
			//Transmission broadcast (0# prefix) in ZigBee (#9 suffix), command light OFF and you want to control the all UNIT : * 1 * 0 * 0#00#9##
			whereStr << pCmd->id * 100 + pCmd->unitcode << ZIGBEE_SUFFIX;
			break;
		default:
			Log(LOG_STATUS, "COpenWebNetUSB unknown command: packettype=%d subtype=%d", packettype, subtype);
			return false;
	}

	// Test packet type
	switch (packettype) {
		case pTypeGeneralSwitch:
			// Test general switch subtype
			switch (subtype) {
				case sSwitchBlindsT1:
				case sSwitchBlindsT2:
					//Blinds/Window command
					who = WHO_AUTOMATION;

					if (pCmd->cmnd == gswitch_sOff)
					{
						what = AUTOMATION_WHAT_DOWN;
					}
					else if (pCmd->cmnd == gswitch_sOn)
					{
						what = AUTOMATION_WHAT_UP;
					}
					else if (pCmd->cmnd == gswitch_sStop)
					{
						what = AUTOMATION_WHAT_STOP;
					}
					break;
				case sSwitchLightT1:
				case sSwitchLightT2:
					//Light/Switch command
					who = WHO_LIGHTING;

					if (pCmd->cmnd == gswitch_sOff)
					{
						what = LIGHTING_WHAT_OFF;
					}
					else if (pCmd->cmnd == gswitch_sOn)
					{
						what = LIGHTING_WHAT_ON;
					}
					else if (pCmd->cmnd == gswitch_sSetLevel)
					{
						// setting level of dimmer
						int level = pCmd->level;

						// Set command based on level value
						if (level == 0)
						{
							what = LIGHTING_WHAT_OFF;
						}
						else if (level == 255)
						{
							what = LIGHTING_WHAT_ON;
						}
						else
						{
							// For dimmers we only allow level 0-99

							if (level<20) {
								what = LIGHTING_WHAT_OFF;
							}
							else if (level<30) {
								what = LIGHTING_WHAT_20;
							}
							else if (level<40) {
								what = LIGHTING_WHAT_30;
							}
							else if (level<50) {
								what = LIGHTING_WHAT_40;
							}
							else if (level<60) {
								what = LIGHTING_WHAT_50;
							}
							else if (level<70) {
								what = LIGHTING_WHAT_60;
							}
							else if (level<80) {
								what = LIGHTING_WHAT_70;
							}
							else if (level<90) {
								what = LIGHTING_WHAT_80;
							}
							else if (level<100) {
								what = LIGHTING_WHAT_90;
							}
							else {
								what = LIGHTING_WHAT_100;
							}
						}
					}

				default:
					break;
			}
			break;

		default:
			Log(LOG_STATUS, "COpenWebNetUSB unknown command: packettype=%d subtype=%d", packettype, subtype);
			return false;
	}

	int used = 1;
	if (!FindDevice(pCmd->id, pCmd->unitcode, subtype, &used)) {
		Log(LOG_ERROR, "command received for unknown device : %d/%s", who, whereStr.str().c_str());
		return false;
	}


	std::stringstream whoStr;
	std::stringstream whatStr;
	whoStr << who;
	whatStr << what;

	std::vector<bt_openwebnet> responses;
	bt_openwebnet request(whoStr.str(), whatStr.str(), whereStr.str(), "");
	if (sendCommand(request, responses))
	{
		if (!responses.empty())
		{
			return responses.at(0).IsOKFrame();
		}
	}

	return true;
}

/**
Find OpenWebNetDevice in DB
**/
bool COpenWebNetUSB::FindDevice(int deviceID, int deviceUnit, int subType, int* used)
{
	std::vector<std::vector<std::string> > result;

	//make device ID
	unsigned char ID1 = (unsigned char)((deviceID & 0xFF000000) >> 24);
	unsigned char ID2 = (unsigned char)((deviceID & 0xFF0000) >> 16);
	unsigned char ID3 = (unsigned char)((deviceID & 0xFF00) >> 8);
	unsigned char ID4 = (unsigned char)(deviceID & 0xFF);

	char szIdx[10];
	sprintf(szIdx, "%02X%02X%02X%02X", ID1, ID2, ID3, ID4);

	if (used != nullptr)
	{
		result = m_sql.safe_query("SELECT ID FROM DeviceStatus WHERE (HardwareID==%d) AND (DeviceID=='%q') AND (Unit == %d) AND (Type==%d) AND (Subtype==%d) and Used == %d",
			m_HwdID, szIdx, deviceUnit, pTypeGeneralSwitch, subType, *used);
	}
	else
	{
		result = m_sql.safe_query("SELECT ID FROM DeviceStatus WHERE (HardwareID==%d) AND (DeviceID=='%q') AND (Unit == %d) AND (Type==%d) AND (Subtype==%d)",
			m_HwdID, szIdx, deviceUnit, pTypeGeneralSwitch, subType);
	}

	if (!result.empty())
	{
		return true;
	}

	return false;
}

bool COpenWebNetUSB::writeRead(const char* command, unsigned int commandSize, bool silent)
{
	std::lock_guard<std::mutex> l(readQueueMutex);

	m_readBufferSize = 0;
	memset(m_readBuffer, 0, OPENWEBNET_SERIAL_BUFFER_SIZE);
	m_bHaveReceived = false;

	if (!isOpen()) {
		if (!silent) {
			Log(LOG_ERROR, "COpenWebNet writeRead error : connection not opened");
		}
		return false;
	}
	AsyncSerial::write(command, commandSize);
	while (!m_bHaveReceived) {
		sleep_milliseconds(SLEEP_READ_TIME);
	}

	return true;
}

bool COpenWebNetUSB::sendCommand(bt_openwebnet& command, std::vector<bt_openwebnet>& response, bool silent)
{
	m_bWriting = true;

	//Try to open the Serial Port
	try
	{
		Log(LOG_STATUS, "Using serial port: %s", m_szSerialPort.c_str());
#ifndef WIN32
		openOnlyBaud(
			m_szSerialPort,
			m_iBaudRate,
			boost::asio::serial_port_base::parity(boost::asio::serial_port_base::parity::none),
			boost::asio::serial_port_base::character_size(8)
			);
#else
		open(
			m_szSerialPort,
			m_iBaudRate,
			boost::asio::serial_port_base::parity(boost::asio::serial_port_base::parity::none),
			boost::asio::serial_port_base::character_size(8)
			);
#endif
	}
	catch (boost::exception & e)
	{
		Log(LOG_ERROR, "Error opening serial port!");
#ifdef _DEBUG
		Log(LOG_ERROR, "-----------------\n%s\n-----------------", boost::diagnostic_information(e).c_str());
#else
		(void)e;
#endif
		m_bWriting = false;
		return false;
	}
	catch (...)
	{
		Log(LOG_ERROR, "Error opening serial port!!!");
		m_bWriting = false;
		return false;
	}

	setReadCallback([this](auto d, auto l) { readCallback(d, l); });
	sOnConnected(this);
	m_bIsStarted = true;


	if (!writeRead(OPENWEBNET_COMMAND_SESSION, static_cast<unsigned int>(strlen(OPENWEBNET_COMMAND_SESSION)), silent)) {
		m_bWriting = false;
		return false;
	}

	std::string responseStr((const char*)m_readBuffer, m_readBufferSize);
	bt_openwebnet responseSession(responseStr);
	Log(LOG_STATUS, "COpenWebNet : sent=%s received=%s", OPENWEBNET_COMMAND_SESSION, responseStr.c_str());

	bool bCheckPassword = false;
	if (!responseSession.IsOKFrame()) {
		if (!silent) {
			Log(LOG_STATUS, "COpenWebNet : failed to begin session, no ACK received (%s)", responseStr.c_str());
		}
		m_bWriting = false;
		return false;
	}

	if (!writeRead(command.m_frameOpen.c_str(), static_cast<unsigned int>(command.m_frameOpen.length()), silent)) {
		m_bWriting = false;
		return false;
	}
	if (!silent) {
		Log(LOG_STATUS, "COpenWebNet : sent=%s received=%s", command.m_frameOpen.c_str(), m_readBuffer);
	}

	if (!ParseData((char*)m_readBuffer, m_readBufferSize, response)) {
		if (!silent) {
			Log(LOG_ERROR, "COpenWebNet : Cannot parse answer : %s", m_readBuffer);
		}
		m_bWriting = false;
		return false;
	}

	if (!response.empty() && response[0].IsPwdFrame()) {
		Log(LOG_STATUS, "COpenWebNet : password required, you have to configure your gateway security settings to allow Domoticz");
	}

	m_bWriting = false;
	return true;
}

bool COpenWebNetUSB::ParseData(char* data, int length, std::vector<bt_openwebnet>& messages)
{
	std::string buffer = std::string(data, length);
	size_t begin = 0;
	size_t end = std::string::npos;
	do {
		end = buffer.find(OPENWEBNET_END_FRAME, begin);
		if (end != std::string::npos) {
			bt_openwebnet message(buffer.substr(begin, end - begin + 2));
			messages.push_back(message);
			begin = end + 2;
		}
	} while (end != std::string::npos);

	return true;
}

void COpenWebNetUSB::readCallback(const char *data, size_t len)
{
	std::lock_guard<std::mutex> l(readQueueMutex);
	if (!m_bIsStarted)
		return;

	if (m_bEnableReceive)
	{
		if (len < OPENWEBNET_SERIAL_BUFFER_SIZE) {
			memcpy(m_readBuffer, data, len);
			m_readBufferSize = static_cast<int>(len);
		}
		else {
			memcpy(m_readBuffer, data, OPENWEBNET_SERIAL_BUFFER_SIZE);
			m_readBufferSize = OPENWEBNET_SERIAL_BUFFER_SIZE;
		}
	}

	m_bHaveReceived = true;
}
